/**
 * This program has been developed by a student from the bachelor Information Science at
 * Utrecht University within the Research Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

import React, { useState } from 'react';
import { useGraphQueryResult } from '../../../../../data-access';
import { MapView, PickingInfo } from '@deck.gl/core/typed';
import DeckGL from '@deck.gl/react/typed';
import { getTooltip } from './Tooltip';
import * as d3 from 'd3';
import { createGraph } from './core';
import { generateBaseLayer } from './layers';

type Props = {
  mapType: number;
  Layer: any;
  nodelink: string;
  isAggregated: boolean;
  isFanned: boolean;
  brushing: boolean;
};

export const NodeLinkMap = ({ mapType = 0, nodelink, Layer }: Props) => {
  const [brushing, setBrushing] = useState<boolean>(false);
  const [clusterRadius, setClusterRadius] = useState<number>(40);
  const [viewport, setViewport] = useState<{ [key: string]: number }>({
    longitude: -102.4,
    latitude: 37.8,
    zoom: 0,
  });

  const graphQueryResult = useGraphQueryResult();

  const graph = React.useMemo(() => {
    return createGraph(graphQueryResult.nodes, graphQueryResult.edges);
  }, [graphQueryResult]);

  React.useEffect(() => {
    setViewport({
      ...graph.calculateCenterPoint(),
      zoom: 0,
    });
  }, [graph]);

  const colorScale: d3.ScaleLinear<string, string> = d3.scaleLinear<string>().domain([0, 6000]).range(['blue', 'red']);

  const layers = [
    generateBaseLayer(mapType),
    new Layer({
      graph: graph,
      colorScale: colorScale,
      brushing: brushing,
      clusterRadius: clusterRadius,
    }),
  ];

  const renderTooltip = (info: PickingInfo) => {
    const { object } = info;
    if (!brushing && object) {
      return getTooltip({ object });
    }
    return undefined;
  };

  if (!graphQueryResult) return <h1>No data</h1>;

  return (
    <div>
      {graph && (
        <DeckGL
          layers={layers}
          views={new MapView({ repeat: true, controller: true })}
          initialViewState={viewport}
          getTooltip={({ object }) => object && renderTooltip(object)}
          onViewStateChange={({ viewState }) => setViewport(viewState)}
        />
      )}
      <div
        style={{
          display: 'flex',
          flexDirection: 'column',
          position: 'absolute',
          top: '20px',
          right: '20px',
          backgroundColor: 'white',
          padding: '30px',
          borderRadius: '4px',
          boxShadow: '0px 2px 4px rgba(0, 0, 0, 0.2)',
        }}
      >
        <h3>Control Panel</h3>
        {nodelink === 'cluster' && (
          <label>
            Cluster radius
            <input
              name="Cluster radius"
              type="range"
              min="10"
              max="100"
              value={clusterRadius}
              onChange={(e) => setClusterRadius(parseInt(e.target.value))}
            />
          </label>
        )}
        <label>
          Brushing
          <input name="Brushing" type="checkbox" checked={brushing} onChange={(e) => setBrushing(e.target.checked)} />
        </label>
      </div>
    </div>
  );
};

export default NodeLinkMap;
