import React, { useEffect, useRef, useState } from 'react';
import { GraphType, LinkType, NodeType } from './types';
import { NLPixi } from './components/NLPixi';
import { parseQueryResult } from './components/query2NL';
import { useImmer } from 'use-immer';
import { ML, setShortestPathSource, setShortestPathTarget } from '../../../data-access/store/mlSlice';
import { VisualizationPropTypes, VISComponentType } from '../../types';
import Input from '@graphpolaris/shared/lib/components/inputs';
import { GraphMetaData } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer, SettingsHeader } from '@graphpolaris/shared/lib/vis/configuration';

export interface NodelinkVisProps {
  layout: string;
  showPopUpOnHover: boolean;
}

const configuration: NodelinkVisProps = {
  layout: 'Force directed',
  showPopUpOnHover: true,
};

export const NodeLinkVis = React.memo(({ data, ml, dispatch }: VisualizationPropTypes) => {
  const ref = useRef<HTMLDivElement>(null);
  const [graph, setGraph] = useImmer<GraphType | undefined>(undefined);
  const [highlightNodes, setHighlightNodes] = useState<NodeType[]>([]);
  const [highlightedLinks, setHighlightedLinks] = useState<LinkType[]>([]);

  useEffect(() => {
    if (data) {
      setGraph(
        parseQueryResult(data, ml, {
          defaultX: (ref.current?.clientWidth || 1000) / 2,
          defaultY: (ref.current?.clientHeight || 1000) / 2,
        }),
      );
    }
  }, [data, ml]);

  const onClickedNode = (node: NodeType, ml: ML) => {
    if (graph) {
      if (ml.shortestPath.enabled) {
        setGraph((draft) => {
          let _node = draft?.nodes.find((n) => n.id === node.id);
          if (!_node) return draft;

          if (!ml.shortestPath.srcNode) {
            _node.isShortestPathSource = true;
            dispatch(setShortestPathSource(node.id));
          } else if (ml.shortestPath.srcNode === node.id) {
            _node.isShortestPathSource = false;
            dispatch(setShortestPathSource(undefined));
          } else if (!ml.shortestPath.trtNode) {
            _node.isShortestPathTarget = true;
            dispatch(setShortestPathTarget(node.id));
          } else if (ml.shortestPath.trtNode === node.id) {
            _node.isShortestPathTarget = false;
            dispatch(setShortestPathTarget(undefined));
          } else {
            _node.isShortestPathSource = true;
            _node.isShortestPathTarget = false;
            dispatch(setShortestPathSource(node.id));
            dispatch(setShortestPathTarget(undefined));
          }
          return draft;
        });
      }
    }
  };

  return (
    <>
      <div className="h-full w-full overflow-hidden" ref={ref}>
        <NLPixi
          graph={graph}
          highlightNodes={highlightNodes}
          highlightedLinks={highlightedLinks}
          onClick={(node, pos) => {
            onClickedNode(node, ml);
          }}
        />
      </div>
    </>
  );
});

const NodelinkSettings = ({
  configuration,
  graph,
  updateSettings,
}: {
  configuration: NodelinkVisProps;
  graph: GraphMetaData;
  updateSettings: (val: any) => void;
}) => {
  return (
    <SettingsContainer>
      <Input
        type="dropdown"
        label="Layout"
        value={configuration.layout}
        options={['Force directed']}
        onChange={(val) => updateSettings({ layout: val })}
      />
      <Input
        type="boolean"
        label="Show pop-up on hover"
        value={configuration.showPopUpOnHover}
        onChange={(val) => updateSettings({ showPopUpOnHover: val })}
      />
    </SettingsContainer>
  );
};

export const NodeLinkComponent: VISComponentType = {
  displayName: 'NodeLinkVis',
  component: NodeLinkVis,
  settings: NodelinkSettings,
  configuration: configuration,
};

export default NodeLinkComponent;
