/**
 * This program has been developed by students from the bachelor Computer Science at
 * Utrecht University within the Software Project course.
 * © Copyright Utrecht University (Department of Information and Computing Sciences)
 */

/* istanbul ignore file */
/* The comment above was added so the code coverage wouldn't count this file towards code coverage.
 * We do not test components/renderfunctions/styling files.
 * See testing plan for more details.*/
import React, { ChangeEventHandler, ReactElement, useEffect, useMemo } from 'react';
import {
  Attribute,
  AttributeNames,
  AttributeOrigin,
  EntitiesFromSchema,
  EntityOrigin,
  NodeOrder,
  PaohvisNodeOrder,
  RelationsFromSchema,
  ValueType,
} from '../Types';
import { Sort } from '@mui/icons-material';
import './MakePaohvisMenu.scss';
import { useImmer } from 'use-immer';
import { useGraphQueryResult, useSchemaGraph } from '@graphpolaris/shared/lib/data-access';
import { calculateAttributesAndRelations, calculateAttributesFromRelation } from '../utils/utils';
import calcEntitiesFromQueryResult from '../utils/CalcEntitiesFromQueryResult';
import { isNodeLinkResult } from '../../../shared/ResultNodeLinkParserUseCase';
import { select } from 'd3';
import { Button } from '../../../../components/buttons';

/** The typing for the props of the Paohvis menu */
type MakePaohvisMenuProps = {
  makePaohvis: (
    entityVertical: string,
    entityHorizontal: string,
    relationName: string,
    isEntityFromRelationFrom: boolean,
    chosenAttribute: Attribute,
    nodeOrder: PaohvisNodeOrder
  ) => void;
};

/** The variables in the state of the PAOHvis menu */
type MakePaohvisMenuState = {
  entityVertical: string;
  entityHorizontal: string;
  relationName: string;
  isEntityVerticalEqualToRelationFrom: boolean;
  attributeNameAndOrigin: string;
  isSelectedAttributeFromEntity: boolean;
  isButtonEnabled: boolean;
  sortOrder: NodeOrder;
  isReverseOrder: boolean;

  entitiesFromSchema: EntitiesFromSchema;
  relationsFromSchema: RelationsFromSchema;
  entitiesFromQueryResult: string[];

  relationNameOptions: string[];
  attributeNameOptions: Record<string, string[]>;
  relationValue: string;
  attributeValue: string;
};

/** React component that renders a menu with input fields for adding a new Paohvis visualisation. */
export const MakePaohvisMenu = (props: MakePaohvisMenuProps) => {
  const graphQueryResult = useGraphQueryResult();
  const schema = useSchemaGraph();
  const [state, setState] = useImmer<MakePaohvisMenuState>({
    entityVertical: '',
    entityHorizontal: '',
    relationName: '',
    isEntityVerticalEqualToRelationFrom: true,
    attributeNameAndOrigin: '',
    isSelectedAttributeFromEntity: false,
    isButtonEnabled: false,
    entitiesFromQueryResult: [],
    sortOrder: NodeOrder.degree,
    isReverseOrder: false,
    entitiesFromSchema: {
      entityNames: [],
      attributesPerEntity: {},
      relationsPerEntity: {},
    },
    relationsFromSchema: {
      relationCollection: [],
      relationNames: {},
      attributesPerRelation: {},
    },
    relationNameOptions: [],
    attributeNameOptions: {},

    relationValue: '?',
    attributeValue: '?',
  });

  useEffect(() => {
    if (state.entityVertical) onChangeEntity(state.entityVertical);
  }, [state.entityVertical, state.entitiesFromSchema]);

  //
  // FUNCTIONS
  //

  /**
   * Called when the entity field is changed.
   * Calculates the `relationNameOptions`, resets the `entityHorizontal`, `relationName` and `attributeName` and sets the new state.
   * It has the value of the entity you clicked on.
   */
  function onChangeEntity(newEntity: string): void {
    console.log(newEntity, state.entitiesFromSchema.relationsPerEntity);

    setState((draft) => {
      draft.relationNameOptions = [];
      const relationNames: string[] = state.entitiesFromSchema.relationsPerEntity[newEntity];

      // check if there are any relations
      if (relationNames) {
        // push all relation options to relationNameOptions
        relationNames.forEach((relation) => {
          let relationSplit = state.relationsFromSchema.relationNames[relation].split(':');

          //check if relation is valid
          if (draft.entitiesFromQueryResult.includes(relationSplit[0]) && draft.entitiesFromQueryResult.includes(relationSplit[1])) {
            // check if relation is selfedge
            if (relationSplit[0] == relationSplit[1]) {
              const relationFrom = `${relation}:${EntityOrigin.from}`;
              const relationTo = `${relation}:${EntityOrigin.to}`;
              if (!draft.relationNameOptions.includes(relationFrom) && !draft.relationNameOptions.includes(relationTo)) {
                draft.relationNameOptions.push(relationFrom);
                draft.relationNameOptions.push(relationTo);
              }
            } else draft.relationNameOptions.push(relation);
          }
        });
        // filter out duplicates
        draft.relationNameOptions = draft.relationNameOptions.filter((n, i) => draft.relationNameOptions.indexOf(n) === i);
      }

      draft.relationValue = '';
      draft.attributeValue = '';

      draft.entityHorizontal = '';
      draft.relationName = draft.relationNameOptions.length > 0 ? draft.relationNameOptions[0] : '';
      draft.isEntityVerticalEqualToRelationFrom = true;
      draft.attributeNameAndOrigin = '';
      draft.isButtonEnabled = false;
      draft.attributeNameOptions = {};
      return draft;
    });
  }

  useEffect(() => {
    console.log(state.relationName, state.relationNameOptions);

    if (state.relationName) onChangeRelationName(state.relationName);
  }, [state.relationName, state.relationNameOptions, state.entityVertical]);
  /**
   * Called when the relationName field is changed.
   * Calculates the possible attribute values, resets the `attributeValue`, sets the `entityHorizontal`.
   * @param {React.ChangeEvent<HTMLInputElement>} event The event that is given by the input field when a change event is fired.
   * It has the value of the relation you clicked on.
   */
  function onChangeRelationName(relationName: string): void {
    console.log('onChangeRelationName');
    setState((draft) => {
      draft.attributeValue = '';
      draft.attributeNameAndOrigin = '';
      draft.isButtonEnabled = false;

      const newRelationSplit = relationName.split(':');
      const newRelation = newRelationSplit[0];

      // This value should always be there
      draft.attributeNameOptions[AttributeOrigin.noAttribute] = [ValueType.noAttribute];

      const relationNames = draft.relationsFromSchema.relationNames[newRelation];
      if (relationNames == undefined) throw new Error('This entity does not exist in the schema.');

      const relationSplit: string[] = relationNames.split(':');
      const entityVertical = draft.entityVertical;

      // check if relation is self edge
      if (relationSplit[0] == relationSplit[1]) {
        const entityOrigin = newRelationSplit[1];

        if (!isValidEntityOrigin(entityOrigin))
          throw new Error(`The entity "${entityVertical}" has an invalid entity origin: ${entityOrigin}.`);

        draft.isEntityVerticalEqualToRelationFrom = entityOrigin == EntityOrigin.from;
        draft.entityHorizontal = relationSplit[0];
      }
      // check if entityVertical is the entity in the 'from' of the relation
      else if (entityVertical == relationSplit[0]) {
        draft.isEntityVerticalEqualToRelationFrom = true;
        draft.entityHorizontal = relationSplit[1];
      }
      // check if entityVertical is the entity in the 'to' of the relation
      else if (entityVertical == relationSplit[1]) {
        draft.isEntityVerticalEqualToRelationFrom = false;
        draft.entityHorizontal = relationSplit[0];
      } else throw new Error(`The relationNames from this.relationsFromSchema for ${newRelation} is invalid`);

      draft.attributeNameOptions[AttributeOrigin.entity] = [];

      if (draft.entitiesFromSchema.attributesPerEntity[draft.entityHorizontal]) {
        let allAttributesOfEntity: AttributeNames = draft.entitiesFromSchema.attributesPerEntity[draft.entityHorizontal];

        let attributeNamesOfEntity: string[] = allAttributesOfEntity.textAttributeNames
          .concat(allAttributesOfEntity.numberAttributeNames)
          .concat(allAttributesOfEntity.boolAttributeNames);
        draft.attributeNameOptions[AttributeOrigin.entity] = attributeNamesOfEntity;
      }

      draft.attributeNameOptions[AttributeOrigin.relation] = [];

      if (draft.relationsFromSchema.attributesPerRelation[newRelation]) {
        let allAttributesOfRelation: AttributeNames = draft.relationsFromSchema.attributesPerRelation[newRelation];

        let attributeNamesOfRelation: string[] = allAttributesOfRelation.textAttributeNames
          .concat(allAttributesOfRelation.numberAttributeNames)
          .concat(allAttributesOfRelation.boolAttributeNames);
        draft.attributeNameOptions[AttributeOrigin.relation] = attributeNamesOfRelation;
      }

      return draft;
    });
  }

  useEffect(() => {
    if (state.attributeValue) onChangeAttributeName(state.attributeValue);
  }, [state.attributeValue, state.relationName]);
  /**
   * Called when the attributeName field is changed.
   * Sets the chosen attribute value, enables the "Make" button and sets the state.
   * @param {React.ChangeEvent<HTMLInputElement>} event The event that is given by the input field when a change event is fired.
   * It has the value of the attributeName you clicked on.
   */
  function onChangeAttributeName(newAttribute: string): void {
    setState((draft) => {
      const isSelectedAttributeFromEntity = newAttribute.split(':')[1] == AttributeOrigin.entity;

      // render changes
      draft.attributeNameAndOrigin = newAttribute;
      draft.isSelectedAttributeFromEntity = isSelectedAttributeFromEntity;
      draft.isButtonEnabled = true;
      return draft;
    });
  }

  useEffect(() => {
    if (state.sortOrder) onChangeSortOrder(state.sortOrder);
  }, [state.sortOrder]);
  /**
   * Called when the sort order field is changed.
   * Sets the state by changing the sort order.
   */
  function onChangeSortOrder(value: string): void {
    const newSortOrder: NodeOrder = value as NodeOrder;
    const nodeOrders: string[] = Object.values(NodeOrder);
    if (nodeOrders.includes(newSortOrder)) {
      unflipReverseIconButton();

      setState((draft) => {
        // render changes
        draft.sortOrder = newSortOrder;
        draft.isReverseOrder = false;
        return draft;
      });
    } else throw new Error(newSortOrder + ' is not a sort order.');
  }

  /**
   * Called when the user clicks the reverse order button.
   * Sets the state and changes the svg of the button.
   * @param {React.ChangeEvent<HTMLInputElement>} event The event that is given by the input field when a change event is fired.
   */
  function onClickReverseOrder(): void {
    //change the svg of the button to be the reversed variant
    setState((draft) => {
      draft.isReverseOrder = !draft.isReverseOrder;
      const isReverseOrder = draft.isReverseOrder;
      switch (draft.sortOrder) {
        case NodeOrder.alphabetical:
          if (isReverseOrder) flipReverseIconButtonLabel();
          else unflipReverseIconButtonLabel();
          break;
        default:
          if (isReverseOrder) flipReverseIconButton();
          else unflipReverseIconButton();
          break;
      }
      return draft;
    });
  }

  /**
   * Called when the user clicks the "Make" button.
   * Checks if all fields are valid before calling `makePaohvis()`.
   */
  function onClickMakeButton(): void {
    const relationName = state.relationName.split(':')[0];

    if (
      state.entitiesFromSchema.entityNames.includes(state.entityVertical) &&
      state.entitiesFromSchema.relationsPerEntity[state.entityVertical].includes(relationName)
    ) {
      const attributeNameAndOrigin: string[] = state.attributeNameAndOrigin.split(':');
      const attributeName: string = attributeNameAndOrigin[0];
      const attributeOrigin: AttributeOrigin = attributeNameAndOrigin[1] as AttributeOrigin;
      const chosenAttribute: Attribute = {
        name: attributeName,
        type: getTypeOfAttribute(attributeName),
        origin: attributeOrigin,
      };

      props.makePaohvis(
        state.entityVertical,
        state.entityHorizontal,
        relationName,
        state.isEntityVerticalEqualToRelationFrom,
        chosenAttribute,
        {
          orderBy: state.sortOrder,
          isReverseOrder: state.isReverseOrder,
        }
      );
    } else {
      setState((draft) => {
        draft.isButtonEnabled = false;
        return draft;
      });
      throw new Error('Error: chosen entity or relation is invalid.');
    }
  }

  /**
   * Gets the type of the value of the specified attribute by .
   * @param attributeName The name of the specified attribute.
   * @returns {ValueType} The type of the value of the specified attribute.
   */
  function getTypeOfAttribute(attributeName: string): ValueType {
    // get the correct AttributeNames for the attribute
    const attributeNames = state.isSelectedAttributeFromEntity
      ? state.entitiesFromSchema.attributesPerEntity[state.entityHorizontal]
      : state.relationsFromSchema.attributesPerRelation[state.relationName.split(':')[0]];

    // look up to which ValueType the attribute belongs
    if (attributeNames.boolAttributeNames.includes(attributeName)) return ValueType.bool;
    if (attributeNames.numberAttributeNames.includes(attributeName)) return ValueType.number;
    if (attributeNames.textAttributeNames.includes(attributeName)) return ValueType.text;
    if (attributeName == ValueType.noAttribute) return ValueType.noAttribute;

    throw new Error('Attribute ' + attributeName + ' does not exist');
  }

  /**
   * Checks if the given string is a valid EntityOrigin.
   * @param entityOrigin string that should be checked.
   * @returns {boolean} returns true if the given string is a valid EntityOrigin.
   */
  function isValidEntityOrigin(entityOrigin: string): boolean {
    return entityOrigin == EntityOrigin.from || entityOrigin == EntityOrigin.to;
  }

  /**
   * Called when the user clicks the reverse order button when `Alphabetical` is selected.
   * Changes the svg of the button to indicate that the order is reversed.
   */
  function flipReverseIconButtonLabel(): void {
    select('.reverseIconButton').select('#reverseButtonLabel').text('Z-A');
  }
  /**
   * Called when the user clicks the reverse order button when `Alphabetical` is selected.
   * Changes the svg of the button back to normal.
   */
  function unflipReverseIconButtonLabel(): void {
    select('.reverseIconButton').select('#reverseButtonLabel').text('A-Z');
  }
  /**
   * Called when the user clicks the reverse order button when `Degree` is selected.
   * Changes the svg of the button to indicate that the order is reversed.
   */
  function flipReverseIconButton(): void {
    select('.reverseIconButton').transition().style('transform', 'scaleY(-1)');
  }
  /**
   * Called when the user clicks the reverse order button when `Degree` is selected.
   * Changes the svg of the button back to normal.
   */
  function unflipReverseIconButton(): void {
    select('.reverseIconButton').transition().style('transform', 'scaleY(1)');
  }

  //
  // REACTIVITY
  //

  useEffect(() => {
    resetConfig();
    setState((draft) => {
      draft.entitiesFromSchema = calculateAttributesAndRelations(schema);
      draft.relationsFromSchema = calculateAttributesFromRelation(schema);
      return draft;
    });
  }, [schema]);

  /** This method filters and makes a new Paohvis table. */
  useEffect(() => {
    if (isNodeLinkResult(graphQueryResult)) {
      resetConfig();
      setState((draft) => {
        draft.entitiesFromQueryResult = calcEntitiesFromQueryResult(graphQueryResult);
        return draft;
      });
    } else {
      console.error('Invalid query result!');
    }
  }, [graphQueryResult]);

  /** This resets the configuration. Should be called when the possible entities and relations change. */
  function resetConfig(): void {
    setState((draft) => {
      draft.entityVertical = '';
      draft.entityHorizontal = '';
      draft.relationName = '';
      draft.isEntityVerticalEqualToRelationFrom = true;
      draft.attributeNameAndOrigin = '';
      draft.isSelectedAttributeFromEntity = false;
      draft.isButtonEnabled = false;
      draft.entitiesFromQueryResult = [];

      draft.relationNameOptions = [];
      draft.attributeNameOptions = {};
      return draft;
    });
  }

  //
  // RENDER
  //

  // Retrieve the possible entity options. If none available, set helper message.
  let entityMenuItems: ReactElement[] = useMemo(() => {
    if (state.entitiesFromQueryResult.length > 0) {
      return state.entitiesFromQueryResult.map((entity) => (
        <option key={entity} value={entity}>
          {entity}
        </option>
      ));
    } else
      return [
        <option key="placeholder" value="" disabled>
          No query data available
        </option>,
      ];
  }, [state.entitiesFromQueryResult]);

  // Retrieve the possible relationName options. If none available, set helper message.
  let relationNameMenuItems: ReactElement[] = useMemo(() => {
    if (state.relationNameOptions.length > 0)
      return state.relationNameOptions.map((relation) => (
        <option key={relation} value={relation}>
          {relation}
        </option>
      ));
    else return [];
  }, [state.relationNameOptions]);

  // Retrieve all the possible attributeName options. If none available, set helper message.
  let attributeNameMenuItems: ReactElement[] = [];
  let attributeNameMenuItemsNoAttribute: ReactElement[] = [];
  let attributeNameMenuItemsEntity: ReactElement[] = [];
  let attributeNameMenuItemsRelation: ReactElement[] = [];
  if (state.attributeNameOptions['Entity'] && state.attributeNameOptions['Relation']) {
    attributeNameMenuItemsNoAttribute = state.attributeNameOptions['No attribute'].map((attribute) => (
      <option key={attribute} value={`${attribute}:NoAttribute`}>
        {attribute}
      </option>
    ));
    attributeNameMenuItemsEntity = state.attributeNameOptions['Entity'].map((attribute) => (
      <option key={`${attribute}:Entity`} value={`${attribute}:Entity`}>
        {`${attribute} : Entity`}
      </option>
    ));
    attributeNameMenuItemsRelation = state.attributeNameOptions['Relation'].map((attribute) => (
      <option key={`${attribute}:Relation`} value={`${attribute}:Relation`}>
        {`${attribute} : Relation`}
      </option>
    ));

    attributeNameMenuItems = attributeNameMenuItemsNoAttribute.concat(attributeNameMenuItemsEntity).concat(attributeNameMenuItemsRelation);
  } else attributeNameMenuItems = [];

  // make sort order menu items
  const sortOrderMenuItems: ReactElement[] = Object.values(NodeOrder).map((nodeOrder) => {
    return (
      <option key={nodeOrder} value={nodeOrder}>
        {nodeOrder}
      </option>
    );
  });

  // make the reverse button
  let reverseIcon: ReactElement;
  switch (state.sortOrder) {
    case NodeOrder.alphabetical:
      reverseIcon = <span id="reverseButtonLabel">A-Z</span>;
      break;
    default:
      reverseIcon = <Sort className={'reverseSortIcon'} />;
      break;
  }

  // return the whole MakePaohvisMenu
  // TODO! fix inputs to select
  return (
    <div className="nav card">
      <div className="card-body flex flex-row overflow-y-auto max-w-[60vw]">
        <select
          className="select"
          id="standard-select-entity"
          value={state.entityVertical}
          onChange={(e) => setState({ ...state, entityVertical: e.target.value })}
        >
          {entityMenuItems}
        </select>
        <select
          className={`select ${relationNameMenuItems.length === 0 ? 'select-disabled' : ''}`}
          id="standard-select-relation"
          value={state.relationName}
          onChange={(e) => setState({ ...state, relationName: e.target.value })}
        >
          {relationNameMenuItems}
        </select>
        <select
          className={`select ${attributeNameMenuItems.length === 0 ? 'select-disabled' : ''}`}
          id="standard-select-attribute"
          value={state.attributeNameAndOrigin}
          onChange={(e) => setState({ ...state, attributeValue: e.target.value })}
        >
          {attributeNameMenuItems}
        </select>
        <select
          className="select"
          id="standard-select-sort-order"
          value={state.sortOrder}
          onChange={(e) => setState({ ...state, sortOrder: e.target.value as NodeOrder })}
        >
          {sortOrderMenuItems}
        </select>
        <button className={'reverseIconButton'} color="inherit" onClick={onClickReverseOrder}>
          {reverseIcon}
        </button>

        <Button label="Make" type="primary" variant="solid" disabled={!state.isButtonEnabled} onClick={onClickMakeButton} />
      </div>
    </div>
  );
};

export default MakePaohvisMenu;
