import { useEffect, useRef, useState } from 'react';
import styles from './paohvis.module.scss';
import {
  Attribute,
  Relation,
  AttributeOrigin,
  EntitiesFromSchema,
  FilterInfo,
  FilterType,
  NodeOrder,
  PaohvisData,
  PaohvisNodeOrder,
  RelationsFromSchema,
  ValueType,
} from './Types';
import { useImmer } from 'use-immer';

import { getWidthOfText } from '../../../schema/schema-utils';
import { useGraphQueryResult, useSchemaGraph } from '../../../data-access';
import { isNodeLinkResult } from '../../shared/ResultNodeLinkParserUseCase';
import { calculateAttributesAndRelations, calculateAttributesFromRelation } from './utils/utils';
import VisConfigPanelComponent from '../../shared/VisConfigPanel/VisConfigPanel';
import { PaohvisFilterComponent } from './components/PaohvisFilterComponent';
import Tooltip from './components/Tooltip';
import { pointer, select, selectAll } from 'd3';
import { HyperEdgeRange } from './components/HyperEdgesRange';
import ToPaohvisDataParserUseCase from './utils/ToPaohvisDataParserUsecase';
import MakePaohvisMenu from './components/MakePaohvisMenu';
import { RowLabelColumn } from './components/RowLabelColumn';
import { VISComponentType, VisualizationPropTypes } from '../../types';
import { GraphMetaData } from '@graphpolaris/shared/lib/data-access/statistics';
import { SettingsContainer, SettingsHeader } from '@graphpolaris/shared/lib/vis/configuration';

type PaohvisViewModelState = {
  rowHeight: number;
  hyperedgeColumnWidth: number;
  gapBetweenRanges: number;

  hyperedgesOnRow: number[][];
  allHyperEdges: number[][];

  entitiesFromSchema: EntitiesFromSchema;
  relationsFromSchema: RelationsFromSchema;
  entityVertical: string;
  entityHorizontal: string;
  chosenRelation: string;
  isEntityVerticalEqualToRelationFrom: boolean;
  nodeOrder: { orderBy: NodeOrder; isReverseOrder: boolean };
  paohvisFilters: { nodeFilters: FilterInfo[]; edgeFilters: FilterInfo[] };
  axisInfo: {
    selectedAttribute: {
      name: string;
      type: ValueType;
      origin: AttributeOrigin;
    };
    relation: { collection: string; from: string; to: string };
    isYAxisEntityEqualToRelationFrom: boolean;
  };
};

export type PaohVisProps = {
  rowHeight: number;
  hyperedgeColumnWidth: number;
  gapBetweenRanges: number;
  data?: PaohvisData;
};

const configuration: PaohVisProps = {
  rowHeight: 30,
  hyperedgeColumnWidth: 20,
  gapBetweenRanges: 5,
};

export const PaohVis = ({ configuration }: VisualizationPropTypes) => {
  const svgRef = useRef<SVGSVGElement>(null);
  const graphQueryResult = useGraphQueryResult();
  const schema = useSchemaGraph();

  const [data, setData] = useState<PaohvisData>({
    rowLabels: [],
    hyperEdgeRanges: [],
    maxRowLabelWidth: 0,
  });
  const [viewModel, setViewModel] = useImmer<PaohvisViewModelState>({
    rowHeight: 30,
    hyperedgeColumnWidth: 20,
    gapBetweenRanges: 5,

    hyperedgesOnRow: [],
    allHyperEdges: [],
    entitiesFromSchema: {
      entityNames: [],
      attributesPerEntity: {},
      relationsPerEntity: {},
    },
    relationsFromSchema: {
      relationCollection: [],
      relationNames: {},
      attributesPerRelation: {},
    },

    entityVertical: '',
    entityHorizontal: '',
    chosenRelation: '',
    isEntityVerticalEqualToRelationFrom: true,
    nodeOrder: { orderBy: NodeOrder.degree, isReverseOrder: false },
    paohvisFilters: { nodeFilters: [], edgeFilters: [] },
    axisInfo: {
      selectedAttribute: {
        name: '',
        type: ValueType.noAttribute,
        origin: AttributeOrigin.noAttribute,
      },
      relation: { collection: '', from: '', to: '' },
      isYAxisEntityEqualToRelationFrom: true,
    },
  });

  // const [state, setState] = useState({
  //   entityVertical: '',
  //   entityHorizontal: '',
  //   relation: '',
  //   isEntityFromRelationFrom: true,
  // });

  //
  // Methods
  //

  /**
   * Filters the current PAOHvis visualization.
   * @param {boolean} isTargetEntity Tells if the the filter you want to remove is applied to an entity or relation.
   * @param {string} entityOrRelationType This is the type of the target.
   * @param {string} attribute This is the chosen attribute of the target you want to filter on.
   * @param {string} predicate This is the chosen predicate of how you want to compare the attribute.
   * @param {string} compareValue This is the chosen value which you want to compare the attribute values to.
   */
  function onClickFilterButton(
    isTargetEntity: boolean,
    entityOrRelationType: string,
    attribute: string,
    predicate: string,
    compareValue: string,
  ): void {
    const attributeName: string = attribute.split(':')[0];
    const attributeType: string = attribute.split(':')[1];
    const lowerCaseCompareValue = compareValue.toLowerCase();
    let compareValueTyped: any;
    // the compareValue must be typed based on the type of the attribute.
    switch (attributeType) {
      case ValueType.number:
        if (!Number.isNaN(Number(compareValue))) compareValueTyped = Number(compareValue);
        else throw new Error('Error: This is not a correct input');
        break;
      case ValueType.bool:
        if (lowerCaseCompareValue == 'true') compareValueTyped = true;
        else if (lowerCaseCompareValue == 'false') compareValueTyped = false;
        else throw new Error('Error: This is not a correct input');
        break;
      default:
        compareValueTyped = compareValue;
        break;
    }
    const newFilter: FilterInfo = {
      targetGroup: entityOrRelationType,
      attributeName: attributeName,
      value: compareValueTyped,
      predicateName: predicate,
    };

    if (isTargetEntity) {
      // The current filter is removed and then the new one is added.
      // TODO: to keep all filters, delete this line and improve config menu to see all filters instead of only the last one.
      setViewModel((draft) => {
        draft.paohvisFilters.nodeFilters.pop();
        draft.paohvisFilters.nodeFilters.push(newFilter);
        return draft;
      });
    } else {
      // The current filter is removed and then the new one is added.
      // TODO: to keep all filters, delete this line and improve config menu to see all filters instead of only the last one.
      setViewModel((draft) => {
        draft.paohvisFilters.edgeFilters.pop();
        draft.paohvisFilters.edgeFilters.push(newFilter);
        return draft;
      });
    }
  }

  /**
   * Resets the current chosen filter.
   * @param {boolean} isTargetEntity Tells if the the filter you want to remove is applied to an entity or relation.
   */
  function onClickResetButton(isTargetEntity: boolean): void {
    setViewModel((draft) => {
      if (isTargetEntity) draft.paohvisFilters.nodeFilters.pop();
      else draft.paohvisFilters.edgeFilters.pop();
      return draft;
    });
  }

  /**
   * SHOULD NOT BE HERE: move to tooltip
   * This calculates a new position based on the current position of the mouse.
   * @param {React.MouseEvent} e This is the position of the mouse.
   */
  function onMouseMoveToolTip(e: React.MouseEvent) {
    select('.tooltip')
      .style('top', pointer(e)[1] - 25 + 'px')
      .style('left', pointer(e)[0] + 5 + 'px');
  }

  /**
   * Handles the visual changes when you enter a certain hyperEdge on hovering.
   * @param colIndex This is the index which states in which column you are hovering.
   * @param nameToShow This is the name of the entity which must be shown when you are hovering on this certain hyperEdge.
   */
  function onMouseEnterHyperEdge(colIndex: number, nameToShow: string): void {
    highlightAndFadeHyperEdges([colIndex]);
    highlightAndFadeRows(new Set(viewModel.allHyperEdges[colIndex]));
    showColName(nameToShow);
  }

  /**
   * Handles the visual changes when you leave a certain hyperEdge on hovering.
   * @param colIndex This is the index which states in which column you were hovering.
   */
  function onMouseLeaveHyperEdge(colIndex: number): void {
    unHighlightAndUnFadeHyperEdges([colIndex]);
    unHighlightAndUnFadeRows(new Set(viewModel.allHyperEdges[colIndex]));
    hideColName();
  }

  /**
   * This makes sure that the correct rows are highlighted and the correct rows are faded.
   * @param {Set<number>} rows This is the set with the numbers of the rows which must be highlighted.
   */
  function highlightAndFadeRows(rows: Set<number>) {
    rows.forEach((row) => highlightRow(row));

    const rowsToFade = [];
    for (let i = 0; i < data.rowLabels.length; i++) if (!rows.has(i)) rowsToFade.push(i);
    rowsToFade.forEach((row) => fadeRow(row));
  }

  /**
   * This makes sure that the correct rows are unhighlighted and the correct rows are unfaded.
   * @param {Set<number>} rows This is the set with the numbers of the rows which must be unhighlighted.
   */
  function unHighlightAndUnFadeRows(rows: Set<number>) {
    rows.forEach((row) => unHighlightRow(row));

    const rowsToUnFade = [];
    for (let i = 0; i < data.rowLabels.length; i++) if (!rows.has(i)) rowsToUnFade.push(i);
    rowsToUnFade.forEach((row) => unFadeRow(row));
  }

  /**
   * This makes sure that the correct hyperEdges are highlighted and the correct rows are faded.
   * @param {number[]} hyperEdges This is the list with the numbers of the hyperEdges which must be highlighted.
   */
  function highlightAndFadeHyperEdges(hyperEdges: number[]) {
    hyperEdges.forEach((hyperEdge) => highlightHyperedge(hyperEdge));

    const colsToFade = [];
    for (let i = 0; i < viewModel.allHyperEdges.length; i++) if (!hyperEdges.includes(i)) colsToFade.push(i);
    colsToFade.forEach((col) => fadeHyperedge(col));
  }

  /**
   * This makes sure that the correct hyperEdges are unhighlighted and the correct rows are unfaded.
   * @param {number[]} hyperEdges This is the list with the numbers of the hyperEdges which must be unhighlighted.
   */
  function unHighlightAndUnFadeHyperEdges(hyperEdges: number[]) {
    hyperEdges.forEach((hyperEdge) => unHighlightHyperedge(hyperEdge));

    const colsToUnFade = [];
    for (let i = 0; i < viewModel.allHyperEdges.length; i++) if (!hyperEdges.includes(i)) colsToUnFade.push(i);
    colsToUnFade.forEach((col) => unFadeHyperedge(col));
  }

  /**
   * This highlights one row given its index.
   * @param {number} rowIndex This is the index of the row that must be highlighted.
   */
  function highlightRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').transition().duration(120).style('font-size', '15px');
  }

  /**
   * This unhighlights one row given its index.
   * @param {number} rowIndex This is the index of the row that must be unhighlighted.
   */
  function unHighlightRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').transition().duration(120).style('font-size', '14px');
  }

  /**
   * This highlights one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be highlighted.
   */
  function highlightHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').transition().duration(120).style('fill', 'orange');
  }

  /**
   * This unhighlights one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be unhighlighted.
   */
  function unHighlightHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').transition().duration(120).style('fill', 'white');
  }

  /**
   * This fades one row given its index.
   * @param {number} rowIndex This is the index of the row that must be faded.
   */
  function fadeRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').attr('opacity', '.3');
  }

  /**
   * This unfades one row given its index.
   * @param {number} rowIndex This is the index of the row that must be unfaded.
   */
  function unFadeRow(rowIndex: number) {
    const row = selectAll('.row-' + rowIndex);
    row.select('text').attr('opacity', '1');
  }

  /**
   * This fades one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be faded.
   */
  function fadeHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').attr('opacity', '.3');
    hyperedge.selectAll('line').attr('opacity', '.3');
  }

  /**
   * This unfades one hyperEdge given its index.
   * @param {number} columnIndex This is the index of the hyperEdge that must be unfaded.
   */
  function unFadeHyperedge(columnIndex: number) {
    const hyperedge = select('.col-' + columnIndex);
    hyperedge.selectAll('circle').attr('opacity', '1');
    hyperedge.selectAll('line').attr('opacity', '1');
  }

  /**
   * This shows the name when hovering on a hyperEdge.
   * @param {string} nameToShow This is the name that must be shown.
   */
  function showColName(nameToShow: string): void {
    select('.tooltip').text(nameToShow).style('visibility', 'visible');
  }

  /** This hides the name when leaving a hyperEdge. */
  function hideColName(): void {
    select('.tooltip').style('visibility', 'hidden');
  }

  /**
   * Makes the new PAOHvis visualisation.
   * @param {string} entityVertical This is the name of the vertical entity (so on the left).
   * @param {string} entityHorizontal This is the name of the horizontal entity (so at the top).
   * @param {string} relationName This is the (collection)-name of the relation.
   * @param {boolean} isEntityVerticalEqualToRelationFrom Tells if the vertical entity is the from or to of the relation.
   * @param {Attribute} chosenAttribute This is the attribute on which the PAOHvis must be grouped by.
   * @param {PaohvisNodeOrder} nodeOrder Defines the sorting order of the PAOHvis visualisation.
   */
  function onClickMakeButton(
    entityVertical: string,
    entityHorizontal: string,
    relationName: string,
    isEntityVerticalEqualToRelationFrom: boolean,
    chosenAttribute: Attribute,
    nodeOrder: PaohvisNodeOrder,
  ): void {
    setViewModel((draft) => {
      draft.entityVertical = entityVertical;
      draft.entityHorizontal = entityHorizontal;
      draft.chosenRelation = relationName;
      draft.isEntityVerticalEqualToRelationFrom = isEntityVerticalEqualToRelationFrom;
      draft.nodeOrder = nodeOrder;

      return draft;
    });
    setAxisInfo(relationName, isEntityVerticalEqualToRelationFrom, chosenAttribute);
  }

  /**
   * This method parses data, makes a new Paohvis Table and lets the view re-render.
   */
  function makePaohvisTable() {
    // set new data
    const newData = new ToPaohvisDataParserUseCase(graphQueryResult).parseQueryResult(viewModel.axisInfo, viewModel.nodeOrder);
    console.log(newData, graphQueryResult, viewModel);

    setData(newData);
  }

  /**
   * This method makes and sets a new PaohvisAxisInfo.
   * @param relationName is the relation that will be used in the Paohvis table.
   * @param isEntityVerticalEqualToRelationFrom is true when the entity on the y-axis belongs to the 'from' part of the relation.
   * @param chosenAttribute is the chosen attribute that will be used to divide the HyperEdges into HyperEdgeRanges.
   */
  function setAxisInfo(relationName: string, isEntityVerticalEqualToRelationFrom: boolean, chosenAttribute: Attribute): void {
    const namesOfEntityTypes = viewModel.relationsFromSchema.relationNames[relationName].split(':');
    const relation: Relation = {
      collection: relationName,
      from: namesOfEntityTypes[0],
      to: namesOfEntityTypes[1],
    };

    // set axisInfo
    setViewModel((draft) => {
      draft.axisInfo = {
        selectedAttribute: chosenAttribute,
        relation: relation,
        isYAxisEntityEqualToRelationFrom: isEntityVerticalEqualToRelationFrom,
      };
      return draft;
    });
  }

  /**
   * Handles the visual changes when you enter a certain row on hovering.
   * @param {number} rowIndex This is the index which states in which row you are hovering.
   */
  function onMouseEnterRow(rowIndex: number): void {
    const rowsToHighlight = new Set<number>();
    const colsToHighlight: number[] = [];
    viewModel.hyperedgesOnRow[rowIndex].forEach((hyperedge) => {
      colsToHighlight.push(hyperedge);

      viewModel.allHyperEdges[hyperedge].forEach((row) => {
        rowsToHighlight.add(row);
      });
    });

    highlightAndFadeRows(rowsToHighlight);
    highlightAndFadeHyperEdges(colsToHighlight);
  }

  /**
   * Handles the visual changes when you leave a certain row on hovering.
   * @param {number} rowIndex This is the index which states in which row you were hovering.
   */
  function onMouseLeaveRow(rowIndex: number): void {
    const rowsToUnHighlight = new Set<number>();
    const colsToUnHighlight: number[] = [];
    viewModel.hyperedgesOnRow[rowIndex].forEach((hyperedge) => {
      colsToUnHighlight.push(hyperedge);

      viewModel.allHyperEdges[hyperedge].forEach((row) => {
        rowsToUnHighlight.add(row);
      });
    });
    unHighlightAndUnFadeRows(rowsToUnHighlight);
    unHighlightAndUnFadeHyperEdges(colsToUnHighlight);
  }

  //
  // Reactivity
  //

  useEffect(() => {
    setViewModel((draft) => {
      // When a schema is received, extract the entity names, and attributes per data type
      draft.entitiesFromSchema = calculateAttributesAndRelations(schema);
      draft.relationsFromSchema = calculateAttributesFromRelation(schema);
      return draft;
    });
  }, [schema]);

  /** This method filters and makes a new Paohvis table. */
  useEffect(() => {
    if (isNodeLinkResult(graphQueryResult)) {
      makePaohvisTable();
    } else {
      console.error('Invalid query result!');
    }
  }, [viewModel.paohvisFilters, graphQueryResult, viewModel.axisInfo, viewModel.nodeOrder]);

  useEffect(() => {
    /**
     * Recalculates the hyperEdges and hyperEdgeRanges for the state.
     */
    setViewModel((draft) => {
      // Create connecteHyperedgesInColumns, to quickly lookup which hyperedges are on a row
      draft.allHyperEdges = [];

      data.hyperEdgeRanges.forEach((hyperEdgeRange) => {
        hyperEdgeRange.hyperEdges.forEach((hyperEdge) => {
          draft.allHyperEdges.push(hyperEdge.indices);
        });
      });

      draft.hyperedgesOnRow = data.rowLabels.map(() => []);
      draft.allHyperEdges.forEach((hyperedge, i) => {
        hyperedge.forEach((row) => {
          draft.hyperedgesOnRow[row].push(i);
        });
      });
      return draft;
    });
  }, [data]);

  //
  // RENDER
  //

  const hyperEdgeRanges = data.hyperEdgeRanges;
  const rowLabelColumnWidth = data.maxRowLabelWidth;
  const hyperedgeColumnWidth = configuration.hyperedgeColumnWidth;

  //calculate yOffset
  let maxColWidth = 0;
  hyperEdgeRanges.forEach((hyperEdgeRange) => {
    const textLength = getWidthOfText(hyperEdgeRange.rangeText, styles.tableFontFamily, styles.tableFontSize, styles.tableFontWeight);
    if (textLength > maxColWidth) maxColWidth = textLength;
  });
  const columnLabelAngleInRadians = Math.PI / 6;
  const yOffset = Math.sin(columnLabelAngleInRadians) * maxColWidth;
  const expandButtonWidth = parseFloat(styles.expandButtonSize.split('px')[0]);
  const margin = 5;

  //calc table width
  let tableWidth = 0;
  let tableWidthWithExtraColumnLabelWidth = 0;
  hyperEdgeRanges.forEach((hyperEdgeRange) => {
    const columnLabelWidth =
      Math.cos(columnLabelAngleInRadians) *
      getWidthOfText(hyperEdgeRange.rangeText, styles.tableFontFamily, styles.tableFontSize, styles.tableFontWeight);
    const columnWidth = hyperEdgeRange.hyperEdges.length * hyperedgeColumnWidth + configuration.gapBetweenRanges * 3;

    tableWidth += columnWidth;

    if (columnLabelWidth > columnWidth) {
      const currentTableWidthWithLabel = tableWidth + columnLabelWidth;
      if (currentTableWidthWithLabel > tableWidthWithExtraColumnLabelWidth)
        tableWidthWithExtraColumnLabelWidth = currentTableWidthWithLabel;
    }
    if (tableWidth > tableWidthWithExtraColumnLabelWidth) tableWidthWithExtraColumnLabelWidth = tableWidth;
  });
  tableWidthWithExtraColumnLabelWidth += rowLabelColumnWidth + expandButtonWidth + margin;

  //make  all hyperEdgeRanges
  let colOffset = 0;
  let hyperEdgeRangeColumns: JSX.Element[] = [];
  hyperEdgeRanges.forEach((hyperEdgeRange, i) => {
    hyperEdgeRangeColumns.push(
      <HyperEdgeRange
        key={i}
        index={i}
        text={hyperEdgeRange.rangeText}
        hyperEdges={hyperEdgeRange.hyperEdges}
        nRows={data.rowLabels.length}
        colOffset={colOffset}
        xOffset={rowLabelColumnWidth}
        yOffset={yOffset}
        rowHeight={configuration.rowHeight}
        hyperedgeColumnWidth={hyperedgeColumnWidth}
        gapBetweenRanges={configuration.gapBetweenRanges}
        onMouseEnter={onMouseEnterHyperEdge}
        onMouseLeave={onMouseLeaveHyperEdge}
      />,
    );
    colOffset += hyperEdgeRange.hyperEdges.length;
  });

  //display message when there is no Paohvis table to display
  let tableMessage = <span></span>;
  let configPanelMessage = <div></div>;
  if (data.rowLabels.length === 0) {
    tableMessage = <div id={styles.tableMessage}> Please choose a valid PAOHvis configuration </div>;
    configPanelMessage = <div id={styles.configPanelMessage}> Please make a PAOHvis table first </div>;
  }

  // returns the whole PAOHvis visualisation panel
  return (
    <div className={styles.container}>
      <MakePaohvisMenu // render the MakePAOHvisMenu
        makePaohvis={onClickMakeButton}
      />
      <div className="w-full h-full">
        <div className="w-full h-full" onMouseMove={onMouseMoveToolTip}>
          {tableMessage}
          <div className="w-full h-full">
            <svg
              ref={svgRef}
              style={{
                width: tableWidthWithExtraColumnLabelWidth,
                height: yOffset + (data.rowLabels.length + 1) * configuration.rowHeight,
              }}
            >
              <RowLabelColumn // render the PAOHvis itself
                onMouseEnter={onMouseEnterRow}
                onMouseLeave={onMouseLeaveRow}
                titles={data.rowLabels}
                width={rowLabelColumnWidth}
                rowHeight={configuration.rowHeight} // viewModel.rowHeight?
                yOffset={yOffset}
              />
              {hyperEdgeRangeColumns}
            </svg>
          </div>
          <Tooltip />
        </div>
        <VisConfigPanelComponent>
          <PaohvisFilterComponent // render the PaohvisFilterComponent with all three different filter options
            axis={FilterType.yaxis}
            filterPaohvis={onClickFilterButton}
            resetFilter={onClickResetButton}
            entityVertical={viewModel.entityVertical}
            entityHorizontal={viewModel.entityHorizontal}
            relationName={viewModel.chosenRelation}
          />
          <PaohvisFilterComponent
            axis={FilterType.xaxis}
            filterPaohvis={onClickFilterButton}
            resetFilter={onClickResetButton}
            entityVertical={viewModel.entityVertical}
            entityHorizontal={viewModel.entityHorizontal}
            relationName={viewModel.chosenRelation}
          />
          <PaohvisFilterComponent
            axis={FilterType.edge}
            filterPaohvis={onClickFilterButton}
            resetFilter={onClickResetButton}
            entityVertical={viewModel.entityVertical}
            entityHorizontal={viewModel.entityHorizontal}
            relationName={viewModel.chosenRelation}
          />
        </VisConfigPanelComponent>
      </div>
    </div>
  );
};

const PaohSettings = ({
  configuration,
  graph,
  updateSettings,
}: {
  configuration: PaohVisProps;
  graph: GraphMetaData;
  updateSettings: (val: any) => void;
}) => {
  return <SettingsContainer>To be implemented</SettingsContainer>;
};

export const PaohVisComponent: VISComponentType = {
  displayName: 'PaohVis',
  component: PaohVis,
  settings: PaohSettings,
  configuration: configuration,
};

export default PaohVisComponent;
